#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "backup_configs"

import sys
import os
import getopt

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
from Utility import print_status, print_status_warning, print_status_error
import HCPClient
import HDIClient
import HDIConfig
import HDIOperation
import FailoverConfig

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import Alert
import HDIConfigArchive
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import get_configs
import backup_sync_common
import getinfo_common

send_alert = False

def backup_configs(conn, work_dir, myid, config):
    ''' Get information with the specified connection and save it to HCP. '''

    log_info("Start backup configuration. host=%s, dir=%s, ID=%s" % (conn.host, work_dir, myid))

    archive = HDIConfigArchive.HDIConfigArchive()

    # Clean up the working directory
    print_status("Preparing ...")
    archive.prepare_save(work_dir)

    # Get node information (to get HDI version)
    print_status("Getting node information ...")
    nodeop = HDIOperation.Node(conn)
    try:
        version = nodeop.get_version()
    except HDIOperation.OperationError, e:
        log_error("Failed to get HDI version of the node. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get HDI version. error=%s" % e.reason, e.code)

    # Get file system information
    print_status("Getting file system settings ...")
    fsallxml = get_configs.get_file_systems(conn)
    archive.save_file_systems(fsallxml)

    # Save synced file system information
    archive.save_synced_file_systems(backup_sync_common.get_synced_dir_path(config.hdi_conf_dir))

    fslist = HDIConfig.parse_filesystems(xml_data=fsallxml)

    # Get history count for each file system
    history_count = {}
    fsop = HDIOperation.FileSystem(conn)
    for fsinfo in fslist:
        count = fsop.get_history_count(fsinfo.fileSystemName)
        history_count[fsinfo.fileSystemName] = count
    archive.save_history_count(history_count)

    # Get CIFS share information
    print_status("Getting CIFS share settings ...")
    cifsop = HDIOperation.CIFSShare(conn)
    for fsinfo in fslist:
        # Check if any CIFS shares exist on the file system
        cifslist = cifsop.get_all(fsname=fsinfo.fileSystemName)
        if len(cifslist) == 0:
            log_info("No CIFS shares on %s." % fsinfo.fileSystemName)
            continue
        tmp_file = "/home/nasroot/cifsbackup.xml"
        cifsop.backup(fsinfo.fileSystemName, tmp_file)
        archive.save_cifs_shares(tmp_file, fsinfo.fileSystemName)
        Utility.remove_file(tmp_file)

    # Get NFS share information
    print_status("Getting NFS share settings ...")
    nfsop = HDIOperation.NFSShare(conn)
    for fsinfo in fslist:
        # Check if any NFS shares exist on the file system
        nfslist = nfsop.get_all(fsname=fsinfo.fileSystemName)
        if len(nfslist) == 0:
            log_info("No NFS shares on %s." % fsinfo.fileSystemName)
            continue
        tmp_file = "/home/nasroot/nfsbackup.xml"
        nfsop.backup(fsinfo.fileSystemName, tmp_file)
        archive.save_nfs_shares(tmp_file, fsinfo.fileSystemName)
        Utility.remove_file(tmp_file)

    # Get migration task information
    print_status("Getting migration task settings ...")
    taskallxml = get_configs.get_mig_tasks(conn)
    archive.save_mig_tasks(taskallxml)

    # Get Large File Transfer settings
    if version >= HDIConfig.LFT_VERSION:
        print_status("Getting Large File Transfer settings ...")
        lftxml = get_configs.get_lft_setting(conn)
        archive.save_lft_setting(lftxml)

    # Get Cache Resident reporting setting
    print_status("Getting Cache Resident settings ...")
    resxml = get_configs.get_cache_resident_reporting(conn)
    archive.save_cache_resident_reporting(resxml)

    # Get network information
    print_status("Getting network settings ...")
    netxml = get_configs.get_network(conn)
    archive.save_network(netxml)

    # Get host name
    print_status("Getting host name ...")
    hostname = get_configs.get_hostname(conn)
    archive.save_hostname(hostname)

    # Get system UUID
    print_status("Getting system UUID ...")
    system_uuid = get_configs.get_system_uuid(conn)
    archive.save_system_uuid(system_uuid)

    # Build archive
    print_status("Building an archive file ...")
    archive_name = "hdiconf_%s_%s_%s.tar.gz" % (myid, hostname, system_uuid)
    archive_path = archive.build(archive_name)

    return archive_path

def upload_configs(hcp_info, archive_path, config):
    log_info("Uploading an archive file to HCP. archive=%s, path=%s" % (archive_path, backup_sync_common.hdiconf_dir_on_hcp))
    hcp_client = HCPClient.HCPClient(hcp_info, "system-backup-data")
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)
    try:
        ret = hcp_client.put(backup_sync_common.hdiconf_dir_on_hcp, from_file=archive_path)
    except HCPClient.ConnectionError, e:
        log_error("Failed to communicate with HCP. error=%s" % str(e))
        raise Exceptions.ConnectionError(e.reason, e.code)

    if ret['code'] != 201:
        if 'X-HCP-ErrorMessage' in ret['headers']:
            errmsg = ret['headers']['X-HCP-ErrorMessage']
        else:
            errmsg = "None"
        log_error("Unexpected HTTP status code. ret=%s, msg=%s, error=%s" % (ret['code'], ret['message'], errmsg))
        log_error("Failed to upload an HDI configuration archive file to HCP. error=%s" % ret['message'])
        raise Exceptions.OperationError("File upload error to HCP. error=%s" % ret['message'], ret['code'])
    log_info("Archive file was uploaded.")

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Parse options
    short_opts = ""
    long_opts = ["send-alert"]
    try:
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError, e:
        raise Exceptions.OptionError(e)

    for o, a in opts:
        if o == "--send-alert":
            global send_alert
            send_alert = True

    # Check if backup is disabled
    if backup_sync_common.is_backup_disabled(config.fo_conf_dir):
        print_status_warning("Executing backup is disabled. Stop execution.")
        raise Exceptions.ConfigurationError("Execution of configuration backup is disabled.")

    # Check resource group status
    print_status("Checking resource group status ...")
    if not getinfo_common.is_running_healthily(config):
        print_status_warning("The node is not running healthily. Stop execution.")
        raise Exceptions.ConfigurationError("Node status is not healthy.")

    # Setup connection information
    # Do not need to specify SSH credential because we are going to make a connection to localhost.
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection("localhost", cred, config.retry_count, config.retry_interval, config.timeout)

    # Get HCP information
    try:
        hcpop = HDIOperation.HCPSetting(conn)
        hcp_info = hcpop.get()
    except HDIOperation.OperationError, e:
        log_error("Failed to get HCP information on the node. host=%s, error=%s" % (conn.host, str(e)))
        raise Exceptions.OperationError("Failed to get HCP settings. error=%s" % e.reason, e.code)

    # Get my identifier
    netop = HDIOperation.Network(conn)
    try:
        iplist = netop.get_iplist()
    except HDIOperation.OperationError, e:
        log_error("Failed to get IP address of this node. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get IP address. error=%s" % e.reason, e.code)
    log_info("My IP addresses: %s" % iplist)
    myid = FailoverConfig.PairConfig.find_matched_ip(config.fo_conf_dir, iplist)
    if myid is None:
        print_status_error("Could not find my IP address in the pair configuration!")
        print_status("Please check the pair configuration if this node exists in the pair configuration.")
        raise Exceptions.ConfigurationError("This node is not defined in the pair configuration.")

    # Backup configs
    archive_path = backup_configs(conn, os.path.join(config.hdi_conf_dir, "backup"), myid, config)

    # Send the archive to HCP
    print_status("Uploading configuration archive file %s to HCP ..." % os.path.basename(archive_path))
    upload_configs(hcp_info, archive_path, config)

    print_status("Completed.")

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        if send_alert:
            Alert.send_alert(Alert.MsgCode.BACKUP_OK)
        sys.exit(0)
    except Exceptions.OptionError, e:
        log_error("Option error: %s" % str(e))
        if send_alert:
            Alert.send_alert(Alert.MsgCode.BACKUP_ERROR, detail=str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        if send_alert:
            Alert.send_alert(Alert.MsgCode.BACKUP_ERROR, detail="Interrupted by a signal.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        if send_alert:
            Alert.send_alert(Alert.MsgCode.BACKUP_ERROR, detail="Unable to load scripts configuration file.")
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Configuration error: %s" % str(e))
        if send_alert:
            Alert.send_alert(Alert.MsgCode.BACKUP_ERROR, detail=str(e))
        sys.exit(20)
    except Exceptions.ConnectionError, e:
        log_error("A connection error occurred: %s" % str(e))
        if send_alert:
            Alert.send_alert(Alert.MsgCode.BACKUP_ERROR, detail=str(e))
        sys.exit(70)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        if send_alert:
            Alert.send_alert(Alert.MsgCode.BACKUP_ERROR, detail=str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        if send_alert:
            Alert.send_alert(Alert.MsgCode.BACKUP_ERROR, detail=str(e))
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        if send_alert:
            Alert.send_alert(Alert.MsgCode.BACKUP_ERROR, detail="Internal error (%s)" % str(e))
        sys.exit(99)

    if send_alert:
        Alert.send_alert(Alert.MsgCode.BACKUP_OK)
    sys.exit(0)
