#!/usr/bin/python

from __future__ import with_statement
import os
import base64

import Utility
import HDIConfig
import HDIClient
import HCPClient
import FailoverConfig

try:
    from CommonLogger import log_debug, log_info, log_warning, log_error
except ImportError:
    # To make this module handy and portable
    def log_debug(msg): pass
    def log_info(msg): pass
    def log_warning(msg): print msg
    def log_error(msg): print msg

# Global variables
# This is used to accelerate getting node information
nodelist = {}

# Class

class BaseError(Exception):
    def __init__(self, reason, code=None):
        self.reason = reason.strip()
        self.code = code

    def __str__(self):
        if self.code is None:
            return "%s" % self.reason
        else:
            return "%s (code=%s)" % (self.reason, self.code)

class ParameterError(BaseError):
    pass

class OperationError(BaseError):
    pass

class ConnectionError(BaseError):
    pass

class ConfigurationError(BaseError):
    pass

class UnsupportedError(BaseError):
    pass

class CommandOperation(object):
    def __init__(self, cmd, secret_cmd=None, display_stdout=False):
        self.cmd = cmd
        self.secret_cmd = secret_cmd
        self.display_stdout = display_stdout

    def execute(self, conn):
        if conn.host == "localhost":
            log_cmd = self.cmd if self.secret_cmd is None else self.secret_cmd
            log_debug("Command-Exec-Start: %s" % log_cmd)
            ret, out, err = Utility.run(self.cmd, self.display_stdout)
            log_debug("Command-Exec-End: %s, ret=%s" % (log_cmd, ret))
        else:
            conn_setting = HDIClient.ConnectionSettings(conn.retry_count, conn.retry_interval, conn.timeout)
            ssh = HDIClient.SSHClient(conn.host, conn.cred, conn_setting)
            try:
                ret, out, err = ssh.exec_cmd(self.cmd, self.secret_cmd, self.display_stdout)
            except HDIClient.ConnectionError, e:
                log_error("Failed to connect an HDI over SSH for command execution. host=%s, code=%s, reason=%s" % (conn.host, e.code, e.reason))
                raise ConnectionError(e.reason, e.code)

        return (ret, out, err)

class RESTOperation(object):
    def __init__(self, method, path, data=None, params=None):
        self.method = method
        self.path = path
        self.data = data
        self.params = params

    def execute(self, conn):
        conn_setting = HDIClient.ConnectionSettings(conn.retry_count, conn.retry_interval, conn.timeout)
        mapi = HDIClient.MAPIClient(conn.host, conn.cred, conn_setting)
        try:
            ret = mapi.exec_method(self.method, self.path, self.data, self.params)
        except HDIClient.ConnectionError, e:
            log_error("Failed to connect to an HDI over HTTP for MAPI execution. host=%s, code=%s, reason=%s" % (conn.host, e.code, e.reason))
            raise ConnectionError(e.reason, e.code)

        return (ret['code'], ret['body'], ret['message'])

class TransferOperation(object):
    def __init__(self, from_path, to_path):
        self.from_path = from_path
        self.to_path = to_path

    def send(self, conn):
        conn_setting = HDIClient.ConnectionSettings(conn.retry_count, conn.retry_interval, conn.timeout)
        ssh = HDIClient.SSHClient(conn.host, conn.cred, conn_setting)
        try:
            ssh.put(self.from_path, self.to_path)
        except HDIClient.ConnectionError, e:
            log_error("Failed to connect an HDI over SSH for file transfer. host=%s, code=%s, reason=%s" % (conn.host, e.code, e.reason))
            raise ConnectionError(e.reason, e.code)

    def receive(self, conn):
        conn_setting = HDIClient.ConnectionSettings(conn.retry_count, conn.retry_interval, conn.timeout)
        ssh = HDIClient.SSHClient(conn.host, conn.cred, conn_setting)
        try:
            ssh.get(self.from_path, self.to_path)
        except HDIClient.ConnectionError, e:
            log_error("Failed to connect an HDI over SSH for file transfer. host=%s, code=%s, reason=%s" % (conn.host, e.code, e.reason))
            raise ConnectionError(e.reason, e.code)

class Connection(object):
    def __init__(self, host, credential, retry_count, retry_interval, timeout):
        self.host = host
        self.cred = credential
        self.retry_count = retry_count
        self.retry_interval = retry_interval
        self.timeout = timeout

class Node(object):
    def __init__(self, conn):
        self.conn = conn
        self.version = None
        self.hostname = None
        self.model = None
        self.uuid = None
        self.number = None
        self.self_mng_ip = None
        self.other_mng_ip = None

    def get_version(self):
        if self.version is None:
            # Get all information
            self._get_info()
        return self.version

    def get_hostname(self):
        if self.hostname is None:
            # Get all information
            self._get_info()
        return self.hostname

    def get_model(self):
        # Return values:
        #   0: Cluster
        #   1: Single/VM
        if self.model is None:
            # Get all information
            self._get_info()
        return self.model

    def get_uuid(self):
        if self.uuid is None:
            # Get all information
            self._get_info()
        return self.uuid

    def get_self_mng_ip(self):
        if self.self_mng_ip is None:
            # Get all information
            self._get_info()
        return self.self_mng_ip

    def get_other_mng_ip(self):
        if self.other_mng_ip is None:
            # Get all information
            self._get_info()
        return self.other_mng_ip

    def get_number(self):
        if self.number is None:
            # Get all information
            self._get_info()
        return self.number

    def is_running_healthily(self):
        # Always run _get_info() to get the latest status.
        log_info("Checking resource group status on %s" % self.conn.host)
        nodeinfo = self._get_info()

        if self.model == HDIConfig.Model.CLUSTER:
            mynode = nodeinfo.PhysicalNodeSelf
            othernode = nodeinfo.PhysicalNodeOther
            for node in (mynode, othernode):
                nodename = node.physicalNodeName
                running = node.runningNode
                rgstatus = node.resourceGroupStatus
                rgerror = node.resourceGroupErrorInfo
                log_info("Node name: %s, Running node: %s, Status: %s, Error info: %s" % (nodename, running, rgstatus, rgerror))
                if rgstatus != "Online":
                    return False
                if rgerror != "No error":
                    return False
                if running != nodename:
                    return False
            return True
        elif self.model == HDIConfig.Model.SINGLE:
            node = nodeinfo.PhysicalNodeSelf
            rgstatus = node.resourceGroupStatus
            rgerror = node.resourceGroupErrorInfo
            log_info("Status: %s, Error info: %s" % (rgstatus, rgerror))
            if rgstatus != "Online":
                return False
            if rgerror != "No error":
                return False
            return True
        else:
            log_error("No model information acquired.")
            return False

    def _get_info(self):
        # Get version, hostname and model
        log_info("Getting node information of %s." % self.conn.host)

        # Issue MAPI
        method = HDIClient.HTTPMethod.GET
        path = "/mapi/ProcessingNode"
        op = RESTOperation(method, path)
        log_info("Acquiring ProcessingNode information from %s." % self.conn.host)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        # Parse the result
        try:
            nodeinfo = HDIConfig.parse_processing_node(xml_data=out)
            version = nodeinfo.PhysicalNodeSelf.SystemVersion['version']
            hostname = nodeinfo.PhysicalNodeSelf.hostName
            model = nodeinfo.SystemConfiguration['mode']
            self_mng_ip = nodeinfo.PhysicalNodeSelf.managementLANAddress
            if nodeinfo.PhysicalNodeOther is not None:
                other_mng_ip = nodeinfo.PhysicalNodeOther.managementLANAddress
            else:
                other_mng_ip = None
            location = nodeinfo.PhysicalNodeSelf.location
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for ProcessingNode. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts
        except (AttributeError, KeyError), e:
            log_error("Expected information is missing in ProcessingNode XML. error=%s" % str(e))
            raise ConfigurationError("Could not get some necessary information for ProcessingNode.")

        # Get system UUID
        cmd = "curl -s -S file:///enas/hsmarc/conf/system_uuid.conf"
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)
        uuid = out

        # Remember all information at once so that we don't need to do this every time.

        # Ex.) "6.4.2-00" -> "6.4.2.00" -> Version("6.4.2.00")
        # version must not be None. Expect the value is always contained in the output of ProcessingNode.
        self.version = Utility.Version(version.replace('-', '.'))
        self.hostname = hostname
        if model == "CLUSTER":
            self.model = HDIConfig.Model.CLUSTER
        elif model == "SINGLE":
            # VM is also presented as "SINGLE"
            self.model = HDIConfig.Model.SINGLE

        self.uuid = uuid

        self.self_mng_ip = self_mng_ip
        self.other_mng_ip = other_mng_ip
        self.number = 1 if location == "node 1" else 0

        log_info("ProcessingNode information: version=%s, hostname=%s, model=%s, uuid=%s, number=%s, self=%s, other=%s" % (self.version, self.hostname, HDIConfig.Model.tostr(self.model), self.uuid, self.number, self.self_mng_ip, self.other_mng_ip))

        # Remember this node information in global list
        global nodelist
        nodelist[self.conn.host] = {'version': self.version, 'hostname': self.hostname, 'model': self.model}

        # Return nodeinfo for individual use
        return nodeinfo

class Storage(object):
    def __init__(self, conn):
        self.conn = conn
        self.model = None
        self.lulist = None      # For simulate mode
        self.vgfreelist = None  # For simulate mode

    def get_lulist(self, empty=False):
        """ Get list of LUs through the connection and return it.

        Args:
            empty (bool): Specify True if you want to get list of empty
                LUs.
        Returns:
            List of HDIConfig.LUInfo: List of LU information.
        """
        log_info("Getting LU information on %s" % self.conn.host)

        model = self._get_model()
        if model == HDIConfig.Model.SINGLE:
            log_error("Wrong operation. LU information acquisition is called for Single node.")
            raise ParameterError("LU information acquisition is called for Single node.")

        method = HDIClient.HTTPMethod.GET
        path = "/mapi/LUs"
        op = RESTOperation(method, path)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        lulist = []
        try:
            tmplulist = HDIConfig.parse_lus(xml_data=out)
            for lu in tmplulist:
                if empty and lu.function != "Free":
                    continue
                lulist.append(lu)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for LUs. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts
        except (AttributeError, KeyError), e:
            log_error("Expected information is missing in LUs XML. error=%s" % str(e))
            raise ConfigurationError("Could not get some necessary information for LUs.")

        return lulist

    def get_vglist(self):
        log_info("Getting volume group information on %s." % self.conn.host)

        model = self._get_model()
        if model == HDIConfig.Model.CLUSTER:
            log_error("Wrong operation. VG information acquisition is called for Cluster node.")
            raise ParameterError("VG information acquisition is called for Cluster node.")

        method = HDIClient.HTTPMethod.GET
        path = "/mapi/VolumeGroups"
        op = RESTOperation(method, path)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        try:
            vglist = HDIConfig.parse_vg(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for VolumeGroups. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts
        except (AttributeError, KeyError), e:
            log_error("Expected information is missing in VolumeGroups XML. error=%s" % str(e))
            raise ConfigurationError("Could not get some necessary information for VolumeGroups.")

        return vglist

    def pickup_lu(self, capacity=None, name=None, refresh=False):
        """ Pick up LU(s) based on the specified condition.

        If capacity is specified, this function picks up LU(s) until total
        capacity reaches the specified capacity.
        If name is specified, this function picks up the LU which has the
        specified name.
        If neither capacity nor name is specified, this function just picks
        up one LU from the top of the list.

        If capacity is specified as 0, this function just returns an empty
        list, with indicating it was able to get sufficient capacity.

        In any case, this function returns a list of LUs even if only one LU
        meets the specified condition.

        If refresh is True, this function retrieves list of empty LUs from
        the specified node (specified by self.conn). Otherwise, this function
        re-uses the last retrieved information.

        Args:
            capacity (int): The least capacity in GB that this function tries
                to acquire from the list of LUs.
            name (str): The name of LU that you want to pick up.
            refresh (bool): If this is True, this function retrieves the latest
                information of empty LU list from the node.
        Returns:
            Returns tuple of the two items below.
            List of LUInfo: List of LUs that meet the specified condition. An
                empty list is returned if no LUs meet the condition.
            bool: True if the requested capacity or LU was successfully acquired.
                False if not.
        """
        log_info("Picking up LU(s) on %s (capacity: %s GB, name: %s)." % (self.conn.host, capacity, name))

        # Retrieves the latest LU information from the node, if refresh is True
        # or if this is the first time execution.
        if refresh or self.lulist is None:
            self.lulist = self.get_lulist(empty=True)
        log_debug("Empty LU list: %s" % [lu.deviceFileName for lu in self.lulist])

        is_sufficient = True
        got_lulist = []
        if name is not None:
            # Pick up the specified LU
            got_lulist = [lu for lu in self.lulist if lu.deviceFileName == name]
            self._remove_lu_by_name(name)
            if len(got_lulist) == 0:
                log_warning("Couldn't get the specified LU. requested=%s" % name)
                is_sufficient = False
        elif capacity == 0:
            # Just return an empty list with is_sufficient = True
            got_lulist = []
            is_sufficient = True
        else:
            # Picks up LUs until total capacity reaches to "capacity"
            lulist = self.lulist[:]     # Copy the list so that we can remove an element in the list in for loop.
            got_capacity = 0
            for lu in lulist:
                got_lulist.append(lu)
                got_capacity += lu.size
                self._remove_lu_by_name(lu.deviceFileName)
                if capacity is None:
                    # Pick up only one LU
                    break
                elif got_capacity >= capacity:
                    # Got enough capacity
                    break
            else:
                log_warning("Couldn't get requested capacity. requested=%s, got=%s" % (capacity, got_capacity))
                is_sufficient = False

        return (got_lulist, is_sufficient)

    def _remove_lu_by_name(self, name):
        idx_to_be_removed = None
        for i, lu in enumerate(self.lulist):
            if lu.deviceFileName == name:
                idx_to_be_removed = i
                break

        if idx_to_be_removed is not None:
            del self.lulist[idx_to_be_removed]
        else:
            log_info("Specified LU (%s) was not removed from LU list because it does not exist in the list." % name)

    def get_free_lulist(self, refresh=False):
        # Retrieves the latest LU information from the node, if refresh is True
        # or if this is the first time execution.
        if refresh or self.lulist is None:
            self.lulist = self.get_lulist(empty=True)

        return self.lulist

    def get_capacity(self, capacity, refresh=False):
        log_info("Getting capacity %s (GB) from volume group on %s." % (capacity, self.conn.host))
        if refresh or self.vgfreelist is None:
            vglist = self.get_vglist()
            vgfreelist = {}
            for vg in vglist:
                vgfreelist[vg.volumeGroupName] = vg.freeSizeInGigaByte
            self.vgfreelist = vgfreelist
        if len(self.vgfreelist) == 0:
            log_error("No volume groups found.")
            raise ConfigurationError("No volume groups found.")

        log_info("VG free list: %s" % self.vgfreelist)
        # Get a VG(s) name with the largest free capacity
        max_free_capacity = max(self.vgfreelist.values())
        max_free_capacity_vglist = [vgname for vgname in self.vgfreelist if self.vgfreelist[vgname] == max_free_capacity]
        # Get a VG whose name comes first alphabetically
        max_free_capacity_vg = sorted(max_free_capacity_vglist)[0]

        is_sufficient = True
        if capacity == 0:
            got_capacity = 0
            is_sufficient = True    # Just to be explicit
        elif capacity <= self.vgfreelist[max_free_capacity_vg]:
            got_capacity = capacity
            self.vgfreelist[max_free_capacity_vg] -= capacity
        else:
            got_capacity = self.vgfreelist[max_free_capacity_vg]
            self.vgfreelist[max_free_capacity_vg] = 0
            log_warning("Couldn't get requested capacity. requested=%s, got=%s, remaining=%s." % (capacity, got_capacity, self.vgfreelist))
            is_sufficient = False

        return (got_capacity, is_sufficient)

    def get_free_capacity(self, refresh=False):
        if refresh or self.vgfreelist is None:
            vglist = self.get_vglist()
            vgfreelist = {}
            for vg in vglist:
                vgfreelist[vg.volumeGroupName] = vg.freeSizeInGigaByte
            self.vgfreelist = vgfreelist
        if len(self.vgfreelist) == 0:
            log_error("No volume groups found.")
            raise ConfigurationError("No volume groups found.")

        return self.vgfreelist

    def _get_model(self):
        if self.model is None:
            self._get_nodeinfo()

        return self.model

    def _get_nodeinfo(self):
        if self.conn.host in nodelist:
            self.model = nodelist[self.conn.host]['model']
        else:
            nodeop = Node(self.conn)
            self.model = nodeop.get_model()

class Network(object):
    def __init__(self, conn):
        self.conn = conn

    def get_iplist(self):
        log_info("Getting IP address information from %s." % self.conn.host)

        method = HDIClient.HTTPMethod.GET
        path = "/mapi/NetworkInterfaces"
        op = RESTOperation(method, path)
        log_info("Acquiring network interface information from %s." % self.conn.host)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        try:
            iplist = HDIConfig.parse_interfaces(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for NetworkInterfaces. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts

        if len(iplist) == 0:
            # This should not happen. Something is wrong.
            log_error("No network interfaces found on the node.")
            raise ConfigurationError("No network interfaces found on the node.")

        return iplist

class FileSystem(object):
    def __init__(self, conn):
        self.conn = conn
        self.version = None
        self.model = None

    def get(self, fsname):
        # Get FileSystemInfo for the specified file system
        log_info("Getting file system information for %s from %s." % (fsname, self.conn.host))

        # Issue MAPI
        method = HDIClient.HTTPMethod.GET
        path = "/mapi/FileSystems/%s" % fsname
        op = RESTOperation(method, path)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret == 404:
            log_warning("The specified file system does not exist on the node. filesystem=%s" % fsname)
            return None
        elif ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        try:
            fsinfo = HDIConfig.parse_filesystem(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for FileSystems. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts

        # Get LFT information separately because some information is missing in MAPI result.
        version = self._get_version()
        if version >= HDIConfig.LFT_VERSION:
            # LFT supported version
            try:
                lftlist = self._get_lft(fsname)
            except OperationError, e:
                log_error("Failed to get Large File Transfer settings of %s. error=%s" % (fsname, str(e)))
                raise

            if len(lftlist) == 0:
                # This should not happen because there must be at least one file system (fsname).
                log_error("No LFT information was found for %s." % fsname)
                raise ConfigurationError("No LFT information was found for %s." % fsname)

            fsinfo.LargeFileTransferSetting = lftlist[0]

        return fsinfo

    def get_all(self):
        # Get a list of FileSystemInfo for each file system
        log_info("Getting all file system information from %s." % self.conn.host)

        # Issue MAPI
        method = HDIClient.HTTPMethod.GET
        path = "/mapi/FileSystems"
        op = RESTOperation(method, path)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        try:
            fslist = HDIConfig.parse_filesystems(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for FileSystems. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts

        # Get LFT information separately because some information is missing in MAPI result.
        version = self._get_version()

        if version >= HDIConfig.LFT_VERSION:
            # LFT supported version
            try:
                lftlist = self._get_lft()
            except OperationError, e:
                log_error("Failed to get Large File Transfer settings of all file systems. error=%s" % str(e))
                raise
            for fsinfo in fslist:
                name = fsinfo.fileSystemName
                lftinfo = [e for e in lftlist if e.fileSystemName == name][0]
                fsinfo.LargeFileTransferSetting = lftinfo

        return fslist

    def get_history_count(self, fsname):
        """ Count directories under .history directory of a file system.

        This function lists up directories under .history directory of a file
        system and returns the number of the directories.

        This function returns 0 if there is no .history directory on the file
        system. This can happen when:
            No migration has happened yet.
            The file system is not mounted.
            The file system does not exist.

        Args:
            fsname (str): File system name.

        Returns:
            int: Number of directories under .history directory of the file system.
        """
        log_info("Getting number of history directories on %s from %s ..." % (fsname, self.conn.host))
        cmd = 'ls /mnt/%s/.history' % fsname
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_info("ls finished with an error. Assumes there is no history directories. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            count = 0
        else:
            count = len(out.split())
        log_debug("Count of history directories: %s" % count)

        return count

    def create(self, fsinfo, policy=None, original_fsinfo=None):
        # Create a file system with MAPI.
        # policy can be omitted, in that case "Like for Like" is applied.
        log_info("Creating a file system %s on %s." % (fsinfo.fileSystemName, self.conn.host))

        # Get version and model of the node where the file system is created
        version = self._get_version()
        model = self._get_model()

        # Calculate required capacity of the file system and work space according to the specified policy
        fslulist = wslulist = fs_capacity = ws_capacity = None

        if policy is None:
            # Behaves same as Like for Like if no policy is specified
            # Create "Like for Like" storage policy for the file system
            policy = FailoverConfig.StoragePolicy(fsinfo.fileSystemName, policy=FailoverConfig.StoragePolicy.LIKEFORLIKE)

        # Get storage operation instance for the connection
        storage = Storage(self.conn)

        if model == HDIConfig.Model.CLUSTER:
            fslulist, wslulist, is_sufficient = self.pickup_lu_based_on_policy(fsinfo, policy, storage, original_fsinfo)
        elif model == HDIConfig.Model.SINGLE:
            fs_capacity, ws_capacity, is_sufficient = self.get_capacity_based_on_policy(fsinfo, policy, storage, original_fsinfo)
        if not is_sufficient:
            log_warning("Capacity for %s may not be sufficient." % fsinfo.fileSystemName)

        # Generate XML
        try:
            is_likeforlike = policy.policy == FailoverConfig.StoragePolicy.LIKEFORLIKE
            xml = fsinfo.toxml(model=model, version=version, fslulist=fslulist, wslulist=wslulist, fs_capacity=fs_capacity, ws_capacity=ws_capacity, is_likeforlike=is_likeforlike)
            log_debug("XML of file system to be created: %s" % xml)
        except HDIConfig.FormatError, e:
            log_error("Insufficient parameters for the file system to build an XML for creation. error=%s" % str(e))
            raise
        except HDIConfig.UnsupportedError, e:
            log_error("Trying to build an XML of an unsupported file system. error=%s" % str(e))
            raise UnsupportedError(e.reason, e.code)

        # Issue MAPI
        method = HDIClient.HTTPMethod.PUT
        path = "/mapi/FileSystems"
        op = RESTOperation(method, path, data=xml)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        log_info("Succeeded to create a file system %s on %s." % (fsinfo.fileSystemName, self.conn.host))

    def mount(self, fsname, mountinfo=None):
        # Mount a file system with MAPI
        log_info("Mounting a file system %s on %s." % (fsname, self.conn.host))

        # Generate XML
        try:
            if mountinfo is None:
                # Just mount the file system with R/W mode.
                xml = "<?xml version='1.0' encoding='UTF-8'?><MountSetting><isMounted>true</isMounted><isReadOnly>false</isReadOnly></MountSetting>"
            else:
                xml = mountinfo.toxml()
            log_debug("XML of mount setting: %s" % xml)
        except HDIConfig.FormatError, e:
            log_error("Insufficient parameters for the file system to build an XML for mounting. error=%s" % str(e))
            raise

        # Issue MAPI
        method = HDIClient.HTTPMethod.POST
        path = "/mapi/FileSystems/%s/MountSetting" % fsname
        op = RESTOperation(method, path, data=xml)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to mount a file system %s on %s." % (fsname, self.conn.host))

    def unmount(self, fsname):
        # Unmount a file system with MAPI
        log_info("Unmounting a file system %s on %s." % (fsname, self.conn.host))

        # Generate XML
        xml = "<?xml version='1.0' encoding='UTF-8'?><MountSetting><isMounted>false</isMounted></MountSetting>"

        # Issue MAPI
        method = HDIClient.HTTPMethod.POST
        path = "/mapi/FileSystems/%s/MountSetting" % fsname
        op = RESTOperation(method, path, data=xml)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to unmount a file system %s on %s." % (fsname, self.conn.host))

    def is_mounted(self, fsname):
        """ Check mount status of a file system.

        Returns True if the specified file system is mounted. False if not
        mounted or the file system does not exist.

        Args:
            fsname (str): File system name.
        Returns:
            bool: True if the file system is mounted. False if not mounted
                or the file system does not exist.
        """
        log_info("Getting mount status of %s on %s." % (fsname, self.conn.host))
        method = HDIClient.HTTPMethod.GET
        path = "/mapi/FileSystems/%s/MountSetting" % fsname
        op = RESTOperation(method, path)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        is_mounted = None
        if ret not in (200, 404):
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)
        elif ret == 404:
            log_info("File system %s does not exist. Return 'Unmounted' as mount status.")
            is_mounted = False
        elif ret == 200:
            # Check mount status xml
            mountinfo = HDIConfig.parse_mountsetting(xml_data=out.strip())
            log_info("Mount status of %s: %s" % (fsname, mountinfo.isMounted))
            is_mounted = mountinfo.isMounted

        return is_mounted

    def delete(self, fsname):
        # Delete a file system with MAPI
        log_info("Deleting a file system %s on %s." % (fsname, self.conn.host))

        # Issue MAPI
        method = HDIClient.HTTPMethod.DELETE
        path = "/mapi/FileSystems/%s" % fsname
        op = RESTOperation(method, path)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret == 404:
            log_warning("The specified file system to delete does not exist. fsname=%s" % fsname)
            # Keep continue as it's already deleted
        elif ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to delete a file system %s on %s." % (fsname, self.conn.host))

    def exists(self, fsname):
        # Check existence of a file system with MAPI
        log_info("Checking existence of a file system %s on %s." % (fsname, self.conn.host))

        # Issue MAPI
        method = HDIClient.HTTPMethod.HEAD
        path = "/mapi/FileSystems/%s" % fsname
        op = RESTOperation(method, path)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret == 200:
            result = True
        elif ret == 404:
            result = False
        else:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to check existence of a file system %s on %s. Existence=%s" % (fsname, self.conn.host, result))

        return result

    def set_lft(self, fsname, lftinfo):
        # Set LFT settings with CLI
        version = self._get_version()
        if version < HDIConfig.LFT_VERSION:
            log_info("Skip LFT setting because the node version is not LFT capable. version=%s" % version)
            return

        if lftinfo is None:
            log_info("Skip LFT setting because no LFT information is provided for %s." % fsname)
            return

        log_info("Configuring Large File Transfer settings for %s on %s." % (fsname, self.conn.host))

        if lftinfo.transferPartMinInMegaBytes is None:
            # transferPartMinInMegaBytes can be None if lftinfo was not created from arclftctl result
            cmd = "sudo arclftctl --file-system %s --%s --lower-limit %s" % (fsname, lftinfo.largeFileTransfer.lower(), lftinfo.lowerLimitInMegaBytes)
        else:
            cmd = "sudo arclftctl --file-system %s --%s --lower-limit %s --transfer-part-min %s" % (fsname, lftinfo.largeFileTransfer.lower(), lftinfo.lowerLimitInMegaBytes, lftinfo.transferPartMinInMegaBytes)

        # Issue CLI
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            if ret == 80:
                log_warning("Command returned with a warning. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            else:
                log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to set Large File Transfer settings of a file system %s on %s." % (fsname, self.conn.host))

    @classmethod
    def pickup_lu_based_on_policy(cls, fsinfo, policy, storage, original_fsinfo=None):
        """ Pick up LUs from the node based on the storage policy.

        This function picks up LUs for both file system and work space
        from the specified storage. Necessary LUs are determined by the
        provided storage policy. And this function returns the picked up
        LUs.
        It is possible that sufficient capacity cannot be acquired due to
        not enough free storage. In that case this function returns LUs
        that it was able to acquire, and returns False which indicates
        'not sufficient capacity was acquired'.

        If AFM is not enabled on the file system, an empty list is returned
        as LUs for work space.

        Args:
            fsinfo (HDIConfig.FileSystemInfo): File system information that
                capacity is acquired for.
            policy (FailoverConfig.StoragePolicy): Storage policy information
                for the file system.
            storage (Storage): Storage operation instance of the node where
                the capacities are acquired.
            original_fsinfo (HDIConfig.FileSystemInfo): File system information
                that is used to determine LUs for failback policy.

        Returns:
            tuple of 3 items below.
                list of HDIConfig.LUInfo: Acquired LUs for file system.
                list of HDIConfig.LUInfo: Acquired LUs for work space.
                bool: True if required capacity was acquired. False if the
                    acquired capacity is not sufficient for the policy.
        """
        is_sufficient = True
        lft_enabled = fsinfo.LargeFileTransferSetting is not None and fsinfo.LargeFileTransferSetting.largeFileTransfer == 'Enable'
        if policy.policy == FailoverConfig.StoragePolicy.AUTOTOTAL:
            # Auto - total
            fslulist, is_suff_fs = storage.pickup_lu(capacity=fsinfo.fileSystemSizeInGigaByte)
            ws_capacity = cls.calc_ws_capacity_from_fslulist(fsinfo, fslulist)
            wslulist, is_suff_ws = storage.pickup_lu(capacity=ws_capacity)
            is_sufficient = is_suff_fs and is_suff_ws
        elif policy.policy == FailoverConfig.StoragePolicy.AUTOUSAGE:
            # Auto - usage
            # Round up (+1)
            fs_capacity = int(fsinfo.usedBlocksInMegaBytes / 1024) + 1 + policy.margin_capacity
            # If LFT is used, need to acquire a minimum capacity that LFT requires.
            if lft_enabled and fs_capacity < HDIConfig.LFT_MINCAPACITY:
                log_info("Calculated capacity (%s GB) of %s is not sufficient for LFT. Assign %s GB." % (fs_capacity, fsinfo.fileSystemName, HDIConfig.LFT_MINCAPACITY))
                fs_capacity = HDIConfig.LFT_MINCAPACITY
            fslulist, is_suff_fs = storage.pickup_lu(capacity=fs_capacity)
            ws_capacity = cls.calc_ws_capacity_from_fslulist(fsinfo, fslulist)
            wslulist, is_suff_ws = storage.pickup_lu(capacity=ws_capacity)
            is_sufficient = is_suff_fs and is_suff_ws
        elif policy.policy == FailoverConfig.StoragePolicy.AUTOSPEC:
            # Auto - specified
            # If LFT is used, need to acquire a minimum capacity that LFT requires.
            if lft_enabled and policy.spec_capacity < HDIConfig.LFT_MINCAPACITY:
                log_info("Specified capacity (%s GB) of %s is not sufficient for LFT. Assign %s GB." % (policy.spec_capacity, fsinfo.fileSystemName, HDIConfig.LFT_MINCAPACITY))
                fs_capacity = HDIConfig.LFT_MINCAPACITY
            else:
                fs_capacity = policy.spec_capacity
            fslulist, is_suff_fs = storage.pickup_lu(capacity=fs_capacity)
            ws_capacity = cls.calc_ws_capacity_from_fslulist(fsinfo, fslulist)
            wslulist, is_suff_ws = storage.pickup_lu(capacity=ws_capacity)
            is_sufficient = is_suff_fs and is_suff_ws
        elif policy.policy == FailoverConfig.StoragePolicy.AUTOPICKUP:
            # Auto - pick up
            # If LFT is used, need to acquire a minimum capacity that LFT requires.
            if lft_enabled:
                fs_capacity = HDIConfig.LFT_MINCAPACITY
            else:
                fs_capacity = None
            fslulist, is_suff_fs = storage.pickup_lu(capacity=fs_capacity)
            if not lft_enabled and 'ActMig' in fsinfo.usage:
                # If only AFM is enabled, just pick up one LU without considering capacity
                ws_capacity = None
            else:
                ws_capacity = cls.calc_ws_capacity_from_fslulist(fsinfo, fslulist)
            wslulist, is_suff_ws = storage.pickup_lu(capacity=ws_capacity)
            is_sufficient = is_suff_fs and is_suff_ws
        elif policy.policy == FailoverConfig.StoragePolicy.LIKEFORLIKE:
            # Like for Like
            fslulist = []
            wslulist = []
            # Pick up the same LUs for file system
            for lu in fsinfo.LUs:
                picked_up_lulist, is_suff_fs = storage.pickup_lu(name=lu.deviceFileName)
                if not is_suff_fs:
                    log_error("LU %s does not exist on the system, though it should exist for Like for Like policy." % lu.deviceFileName)
                    is_sufficient = False
                else:
                    fslulist.extend(picked_up_lulist)
            # Pick up the same LUs for work space
            # Do this pickup only when AFM is enabled. WorkSpaceLUs tag still
            # exists even when AFM is disabled (and deviceFileName tag is empty).
            if 'ActMig' in fsinfo.usage:
                for lu in fsinfo.WorkSpaceLUs:
                    picked_up_lulist, is_suff_ws = storage.pickup_lu(name=lu.deviceFileName)
                    if not is_suff_ws:
                        log_error("LU %s does not exist on the system, though it should exist for Like for Like policy." % lu.deviceFileName)
                        is_sufficient = False
                    else:
                        wslulist.extend(picked_up_lulist)
        elif policy.policy == FailoverConfig.StoragePolicy.MANUAL:
            # Manual
            fslulist = []
            wslulist = []
            # Pick up the specified LUs for file system
            for luname in policy.fslulist:
                picked_up_lulist, is_suff_fs = storage.pickup_lu(name=luname)
                if not is_suff_fs:
                    log_error("LU %s does not exist on the system, though it was specified by Manual policy." % luname)
                    is_sufficient = False
                else:
                    fslulist.extend(picked_up_lulist)
            # Pick up the specified LUs for work space
            for luname in policy.wslulist:
                picked_up_lulist, is_suff_ws = storage.pickup_lu(name=luname)
                if not is_suff_ws:
                    log_error("LU %s does not exist on the system, though it was specified by Manual policy." % luname)
                    is_sufficient = False
                else:
                    wslulist.extend(picked_up_lulist)
        elif policy.policy == FailoverConfig.StoragePolicy.FAILBACK:
            # Failback
            if original_fsinfo is None:
                log_error("Original file system information is not provided for failback policy. name=%s" % fsinfo.fileSystemName)
                raise ParameterError("Original file system information is not provided for failback policy.")
            fslulist = []
            wslulist = []
            # Pick up the same LUs for the original file system
            for lu in original_fsinfo.LUs:
                picked_up_lulist, is_suff_fs = storage.pickup_lu(name=lu.deviceFileName)
                if not is_suff_fs:
                    log_error("LU %s does not exist on the system, though it should exist for Failback policy." % lu.deviceFileName)
                    is_sufficient = False
                else:
                    fslulist.extend(picked_up_lulist)
            # Pick up the same LUs for work space
            # Do this pickup only when AFM is enabled. WorkSpaceLUs tag still
            # exists even when AFM is disabled (and deviceFileName tag is empty).
            if 'ActMig' in original_fsinfo.usage:
                for lu in original_fsinfo.WorkSpaceLUs:
                    picked_up_lulist, is_suff_ws = storage.pickup_lu(name=lu.deviceFileName)
                    if not is_suff_ws:
                        log_error("LU %s does not exist on the system, though it should exist for Failback policy." % lu.deviceFileName)
                        is_sufficient = False
                    else:
                        wslulist.extend(picked_up_lulist)

        return (fslulist, wslulist, is_sufficient)

    @classmethod
    def get_capacity_based_on_policy(cls, fsinfo, policy, storage, original_fsinfo=None):
        """ Get capacity from the node based on the storage policy.

        This function acquires file system capacity and work space capacity
        from the specified storage. File system capacity and work space
        capacity are determined by the provided storage policy. And this
        function returns the acquired capacity.
        It is possible that sufficient capacity cannot be acquired due to
        there is not enough free storage. In that case this function returns
        capacities that it was able to acquire, and returns False which indicates
        'not sufficient capacity was acquired'.

        If AFM is not enabled on the file system, 0 is returned as work space
        capacity.

        Args:
            fsinfo (HDIConfig.FileSystemInfo): File system information that
                capacity is acquired for.
            policy (FailoverConfig.StoragePolicy): Storage policy information
                for the file system.
            storage (Storage): Storage operation instance of the node where
                the capacities are acquired.
            original_fsinfo (HDIConfig.FileSystemInfo): File system information
                that is used to determine capacity for failback policy.

        Returns:
            (int, int, bool):
                int: Acquired file system capacity in GB.
                int: Acquired work space capacity in GB.
                bool: True if required capacity was acquired. False if the
                    acquired capacity is not sufficient for the policy.
        """
        is_sufficient = True
        lft_enabled = fsinfo.LargeFileTransferSetting is not None and fsinfo.LargeFileTransferSetting.largeFileTransfer == 'Enable'
        if policy.policy == FailoverConfig.StoragePolicy.AUTOTOTAL:
            # Auto - total
            fs_capacity, is_suff_fs = storage.get_capacity(capacity=fsinfo.fileSystemSizeInGigaByte)
            ws_capacity, is_suff_ws = storage.get_capacity(capacity=cls.calc_ws_capacity_from_fs_capacity(fsinfo, fs_capacity))
            is_sufficient = is_suff_fs and is_suff_ws
        elif policy.policy == FailoverConfig.StoragePolicy.AUTOUSAGE:
            # Auto - usage
            # Round up (+1)
            capacity_required = int(fsinfo.usedBlocksInMegaBytes / 1024) + 1 + policy.margin_capacity
            if lft_enabled and capacity_required < HDIConfig.LFT_MINCAPACITY:
                log_info("Calculated capacity (%s GB) of %s is not sufficient for LFT. Assign %s GB." % (capacity_required, fsinfo.fileSystemName, HDIConfig.LFT_MINCAPACITY))
                capacity_required = HDIConfig.LFT_MINCAPACITY
            fs_capacity, is_suff_fs = storage.get_capacity(capacity=capacity_required)
            ws_capacity, is_suff_ws = storage.get_capacity(capacity=cls.calc_ws_capacity_from_fs_capacity(fsinfo, fs_capacity))
            is_sufficient = is_suff_fs and is_suff_ws
        elif policy.policy == FailoverConfig.StoragePolicy.AUTOSPEC:
            # Auto - specified
            capacity_required = policy.spec_capacity
            if lft_enabled and capacity_required < HDIConfig.LFT_MINCAPACITY:
                log_info("Specified capacity (%s GB) of %s is not sufficient for LFT. Assign %s GB." % (capacity_required, fsinfo.fileSystemName, HDIConfig.LFT_MINCAPACITY))
                capacity_required = HDIConfig.LFT_MINCAPACITY
            fs_capacity, is_suff_fs = storage.get_capacity(capacity=capacity_required)
            ws_capacity, is_suff_ws = storage.get_capacity(capacity=cls.calc_ws_capacity_from_fs_capacity(fsinfo, fs_capacity))
            is_sufficient = is_suff_fs and is_suff_ws
        elif policy.policy == FailoverConfig.StoragePolicy.LIKEFORLIKE:
            # Like for Like
            fs_capacity, is_suff_fs = storage.get_capacity(capacity=fsinfo.fileSystemSizeInGigaByte)
            ws_capacity, is_suff_ws = storage.get_capacity(capacity=fsinfo.workSpaceSizeInGigaByte) if 'ActMig' in fsinfo.usage else (0, True)
            is_sufficient = is_suff_fs and is_suff_ws
        elif policy.policy == FailoverConfig.StoragePolicy.MANUAL:
            # Manual
            fs_capacity, is_suff_fs = storage.get_capacity(capacity=policy.fs_capacity)
            if policy.ws_capacity is None:
                # Automatic calculation for work space
                ws_capacity, is_suff_ws = storage.get_capacity(capacity=cls.calc_ws_capacity_from_fs_capacity(fsinfo, fs_capacity))
            else:
                ws_capacity, is_suff_ws = storage.get_capacity(capacity=policy.ws_capacity) if 'ActMig' in fsinfo.usage else (0, True)
            is_sufficient = is_suff_fs and is_suff_ws
        elif policy.policy == FailoverConfig.StoragePolicy.FAILBACK:
            # Failback
            if original_fsinfo is None:
                log_error("Original file system information is not provided for failback policy. name=%s" % fsinfo.fileSystemName)
                raise ParameterError("Original file system information is not provided for failback policy.")
            # Use original fsinfo just for storage configuration
            fs_capacity, is_suff_fs = storage.get_capacity(capacity=original_fsinfo.fileSystemSizeInGigaByte)
            ws_capacity, is_suff_ws = storage.get_capacity(capacity=original_fsinfo.workSpaceSizeInGigaByte) if 'ActMig' in fsinfo.usage else (0, True)
            is_sufficient = is_suff_fs and is_suff_ws

        return (fs_capacity, ws_capacity, is_sufficient)

    @staticmethod
    def calc_ws_capacity_from_fslulist(fsinfo, fslulist):
        """ Calculate work space capacity for cluster model based on total LU
        capacity for file system.

        This function calculates capacity of the file system based on the provided
        LU list for the file system, and returns the work space capacity required
        for the file system capacity.
        Work space capacity is based on the spec for cluster model. Also this
        function takes LFT usage on the file system into consideration for the
        calculation.

        If AFM is not enabled on the file system, this function just returns 0.

        Args:
            fsinfo (HDIConfig.FileSystemInfo): File system information for that
                work space capacity is calculated.
            fslulist (List of LUInfo): List of LUs used for the file system
                capacity.

        Returns:
            int: Calculated work space capacity in GB.
        """
        # Return 0 if AFM is not used on the file system
        if 'ActMig' not in fsinfo.usage:
            return 0

        # Calculate total capacity of the LUs (i.e. capacity of the file system)
        fs_capacity = sum([lu.size for lu in fslulist])

        # Check LFT usage
        if fsinfo.LargeFileTransferSetting is not None:
            lft = fsinfo.LargeFileTransferSetting.largeFileTransfer == "Enable"
        else:
            lft = False

        # Calculate work space capacity
        ws_capacity = HDIConfig.FileSystemInfo.calculate_workspace_capacity(HDIConfig.Model.CLUSTER, fs_capacity, lft)
        return ws_capacity

    @staticmethod
    def calc_ws_capacity_from_fs_capacity(fsinfo, fs_capacity):
        """ Calculate work space capacity for single model based on the file
        system capacity.

        This function is same as calc_ws_capacity_from_fslulist() method other
        than the fact that this function uses file system capacity provided in
        the arguments to calculate work space capacity.

        Args:
            fsinfo (HDIConfig.FileSystemInfo): File system information for that
                work space capacity is calculated.
            fs_capacity (int): Capacity of the file system in GB.

        Retruns:
            int: Calculated work space capacity in GB.
        """
        # Return 0 if AFM is not used on the file system
        if 'ActMig' not in fsinfo.usage:
            return 0

        # Check LFT usage
        if fsinfo.LargeFileTransferSetting is not None:
            lft = fsinfo.LargeFileTransferSetting.largeFileTransfer == "Enable"
        else:
            lft = False

        # Calculate work space capacity
        ws_capacity = HDIConfig.FileSystemInfo.calculate_workspace_capacity(HDIConfig.Model.SINGLE, fs_capacity, lft)
        return ws_capacity

    def _get_lft(self, fsname=None):
        if fsname is None:
            cmd = "sudo arclftctl --xml"
            log_info("Getting LFT information for all file systems.")
        else:
            cmd = "sudo arclftctl --xml --file-system %s" % fsname
            log_info("Getting LFT information for file system %s." % fsname)
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)
        try:
            lftlist = HDIConfig.parse_arclftctl(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned by arclftctl. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts

        return lftlist

    def _get_version(self):
        if self.version is None:
            self._get_nodeinfo()

        return self.version

    def _get_model(self):
        if self.model is None:
            self._get_nodeinfo()

        return self.model

    def _get_nodeinfo(self):
        if self.conn.host in nodelist:
            self.version = nodelist[self.conn.host]['version']
            self.model = nodelist[self.conn.host]['model']
        else:
            nodeop = Node(self.conn)
            self.version = nodeop.get_version()
            self.model = nodeop.get_model()

class CIFSShare(object):
    def __init__(self, conn):
        self.conn = conn

    def get(self, cifsname):
        # Get the specified CIFS share information with MAPI
        log_info("Getting CIFS share information of %s from %s." % (cifsname, self.conn.host))

        # Issue MAPI to get all CIFS shares
        method = HDIClient.HTTPMethod.GET
        path = "/mapi/CIFSShares"
        op = RESTOperation(method, path)
        log_info("Acquiring all CIFS share information from %s." % self.conn.host)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        try:
            cifslist = HDIConfig.parse_cifsshares(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for CIFSShares. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts

        # Get the specified CIFS share information
        for cifsinfo in cifslist:
            if cifsinfo.cifsShareName == cifsname:
                return cifsinfo

        # Reach here if not returned in the loop above
        log_warning("The specified CIFS share does not exist on the node. share=%s" % cifsname)
        return None

    def get_all(self, fsname=None):
        # Get all CIFS share information (on the specified file system) with MAPI
        log_info("Getting all CIFS share information%s from %s." % ((" on %s" % fsname) if fsname is not None else "", self.conn.host))

        # Issue MAPI to get all CIFS shares
        method = HDIClient.HTTPMethod.GET
        path = "/mapi/CIFSShares"
        op = RESTOperation(method, path)
        log_info("Acquiring all CIFS share information from %s." % self.conn.host)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        try:
            cifslist = HDIConfig.parse_cifsshares(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for CIFSShares. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts

        if fsname is None:
            # Returns all CIFS share
            return cifslist
        else:
            # Returns CIFS shares on the specified file system
            cifslist_fs = []
            for cifsinfo in cifslist:
                if cifsinfo.filesystem == fsname:
                    cifslist_fs.append(cifsinfo)
            return cifslist_fs

    def delete(self, cifsinfo):
        # Delete a CIFS share with MAPI
        log_info("Deleting a CIFS share %s on %s." % (cifsinfo.cifsShareName, self.conn.host))

        # Issue MAPI
        share64 = base64.b64encode(cifsinfo.cifsShareName)
        dir64 = base64.b64encode(cifsinfo.exportPoint)
        method = HDIClient.HTTPMethod.DELETE
        path = "/mapi/CIFSShares/%s" % share64
        params = {'SHARE-DIRECTORY-NAME': dir64}
        op = RESTOperation(method, path, params=params)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret == 404:
            log_warning("The specified share does not exist. share=%s" % cifsinfo.cifsShareName)
            # Keep continue as it's already deleted
        elif ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to delete a CIFS share %s on %s." % (cifsinfo.cifsShareName, self.conn.host))

    def exists(self, cifsname):
        # Check existence of a CIFS share with MAPI
        log_info("Checking existence of CIFS share %s on %s." % (cifsname, self.conn.host))

        # Issue MAPI to get all CIFS shares
        method = HDIClient.HTTPMethod.GET
        path = "/mapi/CIFSShares"
        op = RESTOperation(method, path)
        log_info("Acquiring all CIFS share information from %s." % self.conn.host)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        try:
            cifslist = HDIConfig.parse_cifsshares(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for CIFSShares. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts

        # Check if the specified CIFS share is included in the cifs list
        result = cifsname in [e.cifsShareName for e in cifslist]

        log_info("Succeeded to check existence of a CIFS share %s on %s. Existence=%s" % (cifsname, self.conn.host, result))
        return result

    def backup(self, fsname, to_file):
        # Backup CIFS share information on the specified file system to a file with a CLI
        log_info("Getting backup of CIFS shares on %s on %s to %s" % (fsname, self.conn.host, to_file))

        # Check the specified file path. It must be a file path on /home/nasroot or just a file name.
        if len(os.path.dirname(to_file)) == 0:
            dirpath = "/home/nasroot"
            filepath = os.path.join(dirpath, to_file)
        else:
            dirpath = os.path.abspath(os.path.dirname(to_file))
            filepath = to_file
        if dirpath != "/home/nasroot":
            log_error("A file path other than on /home/nasroot is specified as cifsbackup output file. path=%s" % to_file)
            raise ParameterError("A file path other than on /home/nasroot is specified as cifsbackup output file.")

        # Delete the file if already exists
        log_info("Deleting the backup file before executing cifsbackup. file=%s" % filepath)
        Utility.remove_file(filepath)

        # Execute cifsbackup command
        filename = os.path.basename(filepath)
        cmd = "sudo cifsbackup %s %s" % (fsname, filename)
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to get backup of CIFS share information.")

    def restore(self, from_file):
        # Restore CIFS shares with a CLI
        log_info("Restoring CIFS shares from %s on %s" % (from_file, self.conn.host))

        # Check the specified file path. It must be a file path on /home/nasroot or just a file name.
        if len(os.path.dirname(from_file)) == 0:
            dirpath = "/home/nasroot"
            filepath = os.path.join(dirpath, from_file)
        else:
            dirpath = os.path.abspath(os.path.dirname(from_file))
            filepath = from_file
        if dirpath != "/home/nasroot":
            log_error("A file path other than on /home/nasroot is specified as CIFS share definition file. path=%s" % from_file)
            raise ParameterError("A file path other than on /home/nasroot is specified as CIFS share definition file.")

        # Check existence of the specified file
        if not os.path.exists(filepath):
            log_error("The specified CIFS share definition file does not exist. path=%s" % filepath)
            raise OperationError("The specified CIFS share definition file does not exist.", -1)

        # Execute cifsrestore command
        filename = os.path.basename(filepath)
        cmd = "sudo cifsrestore %s" % filename
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to restore CIFS shares.")

class NFSShare(object):
    def __init__(self, conn):
        self.conn = conn

    def get(self, nfspath):
        # Get the specified NFS share information with MAPI
        log_info("Acquiring NFS share information of %s from %s." % (nfspath, self.conn.host))

        # Issue MAPI to get all NFS shares
        sharename = os.path.basename(nfspath)
        method = HDIClient.HTTPMethod.GET
        path = "/mapi/NFSShares/%s" % sharename
        params = {'SHARE-DIRECTORY-NAME': nfspath}
        op = RESTOperation(method, path, params=params)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret == 404:
            log_warning("The specified NFS share does not exist on the node. share=%s" % nfspath)
            return None
        elif ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        try:
            nfsinfo = HDIConfig.parse_nfsshare(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for NFSShares. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts

        return nfsinfo

    def get_all(self, fsname=None):
        # Get all NFS share information (on the specified file system) with MAPI
        log_info("Getting all NFS share information%s from %s." % ((" on %s" % fsname) if fsname is not None else "", self.conn.host))

        # Issue MAPI to get all NFS shares
        method = HDIClient.HTTPMethod.GET
        path = "/mapi/NFSShares"
        op = RESTOperation(method, path)
        log_info("Acquiring all NFS share information from %s." % self.conn.host)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        try:
            nfslist = HDIConfig.parse_nfsshares(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for NFSShares. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts

        if fsname is None:
            # Returns all NFS share
            return nfslist
        else:
            # Returns NFS shares on the specified file system
            nfslist_fs = []
            for nfsinfo in nfslist:
                if nfsinfo.filesystem == fsname:
                    nfslist_fs.append(nfsinfo)
            return nfslist_fs

    def delete(self, nfsinfo):
        # Delete a NFS share with MAPI
        log_info("Deleting a NFS share %s on %s." % (nfsinfo.exportPoint, self.conn.host))

        # Issue MAPI
        sharename = os.path.basename(nfsinfo.exportPoint)
        method = HDIClient.HTTPMethod.DELETE
        path = "/mapi/NFSShares/%s" % sharename
        params = {'SHARE-DIRECTORY-NAME': nfsinfo.exportPoint}
        op = RESTOperation(method, path, params=params)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret == 404:
            log_warning("The specified share does not exist. share=%s" % nfsinfo.exportPoint)
            # Keep continue as it's already deleted
        elif ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to delete a NFS share %s on %s." % (nfsinfo.exportPoint, self.conn.host))

    def exists(self, nfspath):
        # Check existence of a NFS share with MAPI
        log_info("Checking existence of a NFS share %s on %s." % (nfspath, self.conn.host))

        # Issue MAPI
        sharename = os.path.basename(nfspath)
        method = HDIClient.HTTPMethod.HEAD
        path = "/mapi/NFSShares/%s" % sharename
        params = {'SHARE-DIRECTORY-NAME': nfspath}
        op = RESTOperation(method, path, params=params)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret == 200:
            result = True
        elif ret == 404:
            result = False
        else:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to check existence of a NFS share %s on %s. Existence=%s" % (nfspath, self.conn.host, result))

        return result

    def backup(self, fsname, to_file):
        # Backup NFS share information on the specified file system to a file with a CLI
        log_info("Getting backup of NFS shares on %s on %s to %s" % (fsname, self.conn.host, to_file))

        # Check the specified file path. It must be a file path on /home/nasroot or just a file name.
        dirpath = os.path.abspath(os.path.dirname(to_file))
        if len(dirpath) != 0 and dirpath != "/home/nasroot":
            log_error("A file path other than on /home/nasroot is specified as nfsbackup output file. path=%s" % to_file)
            raise ParameterError("A file path other than on /home/nasroot is specified as nfsbackup output file.")

        # Delete the file if already exists
        filepath = to_file if len(dirpath) != 0 else os.path.join("/home/nasroot", to_file)
        log_info("Deleting the backup file before executing nfsbackup. file=%s" % filepath)
        Utility.remove_file(filepath)

        # Execute nfsbackup command
        filename = os.path.basename(to_file)
        cmd = "sudo nfsbackup %s %s" % (fsname, filename)
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to get backup of NFS share information.")

    def restore(self, from_file):
        # Restore NFS shares with a CLI
        log_info("Restoring NFS shares from %s on %s" % (from_file, self.conn.host))

        # Check the specified file path. It must be a file path on /home/nasroot or just a file name.
        dirpath = os.path.abspath(os.path.dirname(from_file))
        if len(dirpath) != 0 and dirpath != "/home/nasroot":
            log_error("A file path other than on /home/nasroot is specified as NFS share definition file. path=%s" % from_file)
            raise ParameterError("A file path other than on /home/nasroot is specified as NFS share definition file.")

        # Check existence of the specified file
        filepath = from_file if len(dirpath) != 0 else os.path.join("/home/nasroot", from_file)
        if not os.path.exists(filepath):
            log_error("The specified NFS share definition file does not exist. path=%s" % filepath)
            raise OperationError("The specified NFS share definition file does not exist.", -1)

        # Execute nfsrestore command
        filename = os.path.basename(from_file)
        cmd = "sudo nfsrestore %s" % filename
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to restore NFS shares.")

class MigrationTask(object):
    def __init__(self, conn):
        self.conn = conn

    def get(self, taskname, fsname):
        # Get migration task information with MAPI
        log_info("Getting migration task information %s on %s from %s." % (taskname, fsname, self.conn.host))

        # Issue MAPI
        taskname64 = base64.b64encode(taskname)
        method = HDIClient.HTTPMethod.GET
        path = "/mapi/MigrationTask/%s" % taskname64
        params = {'filesystemName': fsname}
        op = RESTOperation(method, path, params=params)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret == 404:
            log_warning("The specified migration task does not exist on the node. task=%s, filesystem=%s" % (taskname, fsname))
            return None
        elif ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        try:
            taskinfo = HDIConfig.parse_migtask(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for MigrationTask. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts

        return taskinfo

    def get_all(self, fsname=None):
        # Get all migration task information with MAPI
        log_info("Getting all migration task information on %s from %s." % (fsname, self.conn.host))

        # Issue MAPI
        method = HDIClient.HTTPMethod.GET
        path = "/mapi/MigrationTask"
        op = RESTOperation(method, path)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        try:
            tasklist = HDIConfig.parse_migtasks(xml_data=out)
        except HDIConfig.FormatError, e:
            log_error("An invalid XML was returned for MigrationTask. error=%s" % str(e))
            raise   # Pass through - this should be a bug of HDI or the scripts

        if fsname is None:
            return tasklist
        else:
            tasklist_fs = []
            for task in tasklist:
                if task.fileSystemName == fsname:
                    tasklist_fs.append(task)
            return tasklist_fs

    def create(self, taskinfo, start_time):
        # Create a migration task with MAPI
        log_info("Creating a migration task %s on %s on %s, starting at %s." % (taskinfo.task_name, taskinfo.fileSystemName, self.conn.host, Utility.epoch_to_iso8601(start_time)))

        # Generate XML
        try:
            xml = taskinfo.toxml(start_time)
            log_debug("XML of a migration task to be created: %s" % xml)
        except HDIConfig.FormatError, e:
            log_error("Insufficient parameters for the migration task to build an XML for creation. error=%s" % str(e))
            raise

        # Issue MAPI
        method = HDIClient.HTTPMethod.PUT
        path = "/mapi/MigrationTask"
        op = RESTOperation(method, path, data=xml)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        log_info("Succeeded to create a migration task %s on %s on %s." % (taskinfo.task_name, taskinfo.fileSystemName, self.conn.host))

    def delete(self, taskname, fsname):
        # Delete a migration task on a file system with MAPI
        log_info("Deleting a migration task %s on file system %s on %s." % (taskname, fsname, self.conn.host))

        # Issue MAPI
        taskname64 = base64.b64encode(taskname)
        method = HDIClient.HTTPMethod.DELETE
        path = "/mapi/MigrationTask/%s" % taskname64
        params = {'filesystemName': fsname}
        op = RESTOperation(method, path, params=params)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret == 404:
            log_warning("The specified migration task does not exist. taskname=%s, fsname=%s" % (taskname, fsname))
            # Keep continue as it's already deleted
        elif ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to delete a migration task %s on %s on %s." % (taskname, fsname, self.conn.host))

    def exists(self, taskname, fsname):
        # Check existence of a migration on a file system with MAPI
        log_info("Checking existence of a migration task %s on a file system %s on %s." % (taskname, fsname, self.conn.host))

        # Issue MAPI
        taskname64 = base64.b64encode(taskname)
        method = HDIClient.HTTPMethod.HEAD
        path = "/mapi/MigrationTask/%s" % taskname64
        params = {'filesystemName': fsname}
        op = RESTOperation(method, path, params=params)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret == 200:
            result = True
        elif ret == 404:
            result = False
        elif ret == 503:
            # It looks HDI returns 503 if the specified task does not exist.
            # Handle 503 same as 404 until it is fixed..
            result = False
        else:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to check existence of a migration task %s on %s on %s." % (taskname, fsname, self.conn.host))

        return result

    def enable(self, taskname, fsname):
        # Enable a migration task on a file system with MAPI
        log_info("Enabling a migration task %s on a file system %s on %s." % (taskname, fsname, self.conn.host))

        # Build an XML
        xml = "<?xml version='1.0' encoding='UTF-8'?><MigrationTask><task-enabled>Enabled</task-enabled><fileSystemName>%s</fileSystemName></MigrationTask>" % fsname

        # Issue MAPI
        taskname64 = base64.b64encode(taskname)
        method = HDIClient.HTTPMethod.POST
        path = "/mapi/MigrationTask/%s" % taskname64
        op = RESTOperation(method, path, data=xml)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        log_info("Succeeded to enable a migration task %s on %s on %s." % (taskname, fsname, self.conn.host))

    def disable(self, taskname, fsname):
        # Disable a migration task on a file system with MAPI
        log_info("Disabling a migration task %s on a file system %s on %s." % (taskname, fsname, self.conn.host))

        # Build an XML
        xml = "<?xml version='1.0' encoding='UTF-8'?><MigrationTask><task-enabled>Disabled</task-enabled><fileSystemName>%s</fileSystemName></MigrationTask>" % fsname

        # Issue MAPI
        taskname64 = base64.b64encode(taskname)
        method = HDIClient.HTTPMethod.POST
        path = "/mapi/MigrationTask/%s" % taskname64
        op = RESTOperation(method, path, data=xml)
        ret, out, err = op.execute(self.conn)
        log_debug("MAPI result: ret=%s, out=%s, err=%s" % (ret, out, err))

        if ret != 200:
            log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s, host=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path, self.conn.host))
            raise OperationError(err, ret)

        log_info("Succeeded to disable a migration task %s on %s on %s." % (taskname, fsname, self.conn.host))

    def start(self, taskname, fsname):
        # Start a migration task on a file system with a CLI
        log_info("Starting a migration task %s on a file system %s on %s." % (taskname, fsname, self.conn.host))

        # Execute CLI
        cmd = "sudo arcmigrun --file-system %s --task %s" % (fsname, taskname)
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to start a migration task %s on %s on %s." % (taskname, fsname, self.conn.host))

    def stop(self, taskname, fsname):
        # Stop a migration task on a file system with a CLI
        log_info("Stopping a migration task %s on a file system %s on %s." % (taskname, fsname, self.conn.host))

        # Execute CLI
        cmd = "sudo arcmigstop --file-system %s --task %s" % (fsname, taskname)
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to stop a migration task %s on %s on %s." % (taskname, fsname, self.conn.host))

class CacheResident(object):
    def __init__(self, conn):
        self.conn = conn

    def is_reporting_enabled(self):
        # Check if reporting is enabled with CLI
        log_info("Checking cache resident reporting setting on %s." % self.conn.host)

        # Execute CLI
        cmd = "sudo arcresidentctl --xml"
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)

        # Parse the result
        enabled = HDIConfig.parse_resident(xml_data=out)
        log_info("Succeeded to get cache resident reporting setting on %s. Enabled=%s" % (self.conn.host, enabled))

        return enabled

    def enable_reporting(self):
        # Enable cache resident reporting with CLI
        log_info("Enabling cache resident reporting on %s." % self.conn.host)

        cmd = "sudo arcresidentctl --report enable"
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to enable cache resident reporting on %s." % self.conn.host)

    def disable_reporting(self):
        # Disable cache resident reporting with CLI
        log_info("Disabling cache resident reporting on %s." % self.conn.host)

        cmd = "sudo arcresidentctl --report disable"
        op = CommandOperation(cmd)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)
        log_info("Succeeded to disable cache resident reporting on %s." % self.conn.host)

class HCPSetting(object):
    def __init__(self, conn):
        self.conn = conn

    def get(self):
        # Get HCP settings from configuration files.
        # Returns HCP settings as HCPClient.HCPInfo format
        # localhost only. Remote host is not supported (not implemented)!
        log_info("Getting HCP settings on %s." % self.conn.host)

        if self.conn.host != "localhost":
            log_error("Getting HCP settings from a remote host is not supported. host=%s" % self.conn.host)
            raise UnsupportedError("Getting HCP settings from a remote host is not supported.")

        hcp_info = HCPClient.HCPInfo()
        try:
            hcp_info.load()
        except HCPClient.FormatError, e:
            log_error("Wrong HCP configuration was detected. error=%s" % str(e))
            raise

        return hcp_info

    def swap_with_replica(self):
        # Swap primary HCP and replica HCP.
        # Returns HCPInfo after swap.
        # This does not execute HCP failover. Just swap HCP information on HDI.
        # localhost only. Remote host is not supported (not implemented)!
        log_info("Swapping primary HCP and remote HCP setting.")

        if self.conn.host != "localhost":
            log_error("Getting HCP settings from a remote host is not supported. host=%s" % self.conn.host)
            raise UnsupportedError("Getting HCP settings from a remote host is not supported.")

        # Get the current HCP settings
        try:
            hcp_info = self.get()
        except OperationError, e:
            log_error("Failed to get HCP settings. error=%s" % str(e))
            raise
        log_info("Current settings: Primary=%s, replica=%s" % (hcp_info.domain, hcp_info.rephost))

        if hcp_info.rephost is None:
            # Replica HCP is not defined.
            log_info("Replica HCP is not defined. Do nothing.")
            return hcp_info

        # Swap Primary and Replica with a CLI
        options = "--host %s --tenant %s --user-name %s --passwordMD5 %s --replica-host %s" % (hcp_info.rephost, hcp_info.tenant, base64.b64decode(hcp_info.user64), hcp_info.passMD5, hcp_info.domain)
        options_nopass = "--host %s --tenant %s --user-name %s --passwordMD5 *** --replica-host %s" % (hcp_info.rephost, hcp_info.tenant, base64.b64decode(hcp_info.user64), hcp_info.domain)
        # Swap external host as well if it is defined
        # Take care of password string for logging
        if hcp_info.exhost is not None:
            if hcp_info.repexhost is not None:
                options += " --ex-host %s --ex-replica-host %s" % (hcp_info.repexhost, hcp_info.exhost)
                options_nopass += " --ex-host %s --ex-replica-host %s" % (hcp_info.repexhost, hcp_info.exhost)
            else:
                # Keep using the current external host if no external replica host is defined
                options += " --ex-host %s" % hcp_info.exhost
                options_nopass += " --ex-host %s" % hcp_info.exhost

        log_info("Updating HCP settings ...")
        cmd = "sudo archcpset %s" % options
        cmd_nopass = "sudo archcpset %s" % options_nopass
        op = CommandOperation(cmd, secret_cmd=cmd_nopass)
        ret, out, err = op.execute(self.conn)
        log_debug("Command result: ret=%s, out=%s, err=%s" % (ret, out, err))
        if ret != 0:
            log_error("A command finished with an error. ret=%s, out=%s, err=%s, cmd=%s, host=%s" % (ret, out, err, cmd, self.conn.host))
            raise OperationError(err, ret)

        # Get the updated HCP settings
        try:
            hcp_info = self.get()
        except OperationError, e:
            log_error("Failed to get HCP settings. error=%s" % str(e))
            raise
        log_info("Succeeded to swap HCP settings on %s. Primary=%s, replica=%s" % (self.conn.host, hcp_info.domain, hcp_info.rephost))

        return hcp_info
