#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"

import sys
import os

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import HDIConfig
import HDIClient
import HDIOperation
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception

sys.path.append(os.path.join(BASEDIR, "bin"))
import Exceptions
import get_configs

# Functions

def get_remote_file_systems(ipaddr, config, ssh_config):
    """ Get a list of file systems on the node specified by ipaddr.

    Args:
        ipaddr (str): IP address of a node which file system information is
            got from.
        config (Config.Config): Configuration of the scripts.
        ssh_config (Config.SSHConfig): SSH configuration.

    Returns:
        List of HDIConfig.FileSystemInfo: List of file system information
            acquired on the node.
    """
    print
    print "Getting a list of file systems on %s ..." % ipaddr

    # Doesn't specify REST credential as it is not used here
    cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(ipaddr), None)
    conn = HDIOperation.Connection(ipaddr, cred, config.retry_count, config.retry_interval, config.timeout)
    fsxml = get_configs.get_file_systems(conn)
    try:
        fs_list = HDIConfig.parse_filesystems(xml_data=fsxml)
    except HDIConfig.FormatError, e:
        raise Exceptions.OperationError("Wrong file system information was returned: %s" % str(e))

    log_debug("File systems on %s: %s" % (ipaddr, [fs.fileSystemName for fs in fs_list]))
    return fs_list

def get_local_file_systems(config):
    """ Get a list of file systems on the local node.
    """

    # No SSH credential is required. Only MAPI is used to get local file systems.
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    fsop = HDIOperation.FileSystem(conn)
    try:
        fs_list = fsop.get_all()
    except HDIOperation.OperationError, e:
        raise Exceptions.OperationError("Failed to get file system information on this node: %s" % str(e))

    return fs_list

def get_local_nodeinfo(config):
    """ Get processing node information of the local node.

    Args:
        config (Config.Config): Configuration of the scripts.

    Returns:
        HDIConfig.ProcessingNodeInfo: Processing node information of the node.
    """
    # No SSH credential is required. Only MAPI is used to get local file systems.
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    method = HDIClient.HTTPMethod.GET
    path = "/mapi/ProcessingNode"
    op = HDIOperation.RESTOperation(method, path)
    log_info("Acquiring ProcessingNode information on the node...")
    try:
        ret, out, err = op.execute(conn)
    except HDIOperation.OperationError, e:
        log_error("Failed to get ProcessingNode information on the node. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get ProcessingNode information on the node. error=%s" % str(e), e.code)

    if ret != 200:
        log_error("Unexpected HTTP status code. ret=%s, out=%s, err=%s, method=%s, path=%s" % (ret, out, err, HDIClient.HTTPMethod.tostr(method), path))
        raise Exceptions.OperationError(err, ret)

    try:
        nodeinfo = HDIConfig.parse_processing_node(xml_data=out)
    except HDIConfig.FormatError, e:
        log_error("An invalid XML was returned for ProcessingNode. error=%s" % str(e))
        raise

    return nodeinfo

def get_remote_model(ipaddr, config, ssh_config):
    """ Get model information of the node specified by ipaddr.

    Args:
        ipaddr (str): IP address of a node which model information is
            got.
        config (Config.Config): Configuration of the scripts.
        ssh_config (Config.SSHConfig): SSH configuration.

    Returns:
        int: A constant number for a model.
    """
    print
    print "Getting model information of %s ..." % ipaddr

    # Doesn't specify REST credential as it is not used here
    cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(ipaddr), None)
    conn = HDIOperation.Connection(ipaddr, cred, config.retry_count, config.retry_interval, config.timeout)
    prog = os.path.join(BASEDIR, "int/get_model")
    cmd = "sudo maintexec %s" % prog
    op = HDIOperation.CommandOperation(cmd)

    try:
        ret, out, err = op.execute(conn)
    except HDIOperation.ConnectionError, e:
        log_error("Failed to communicate with %s. error=%s" % (conn.host, str(e)))
        raise Exceptions.ConnectionError(e.reason, e.code)

    if ret == 0:
        model = HDIConfig.Model.CLUSTER
    elif ret == 1:
        model = HDIConfig.Model.SINGLE
    else:
        log_error("Unknown model information is returned from %s. ret=%s" % (conn.host, ret))
        raise Exceptions.OperationError("Unkown model information.", ret)

    log_debug("Model of %s: %s" % (ipaddr, HDIConfig.Model.tostr(model)))
    return model

def get_local_model(config):
    """ Get model information of localhost.
    """
    # Doesn't need to specify credentials because this is a command operation
    # on localhost.
    cred = HDIClient.Credential(None, None, None, None)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)
    prog = os.path.join(BASEDIR, "int/get_model")
    cmd = "sudo maintexec %s" % prog
    op = HDIOperation.CommandOperation(cmd)

    ret, out, err = op.execute(conn)

    if ret == 0:
        model = HDIConfig.Model.CLUSTER
    elif ret == 1:
        model = HDIConfig.Model.SINGLE
    else:
        log_error("Unknown model information is returned from %s. ret=%s" % (conn.host, ret))
        raise Exceptions.OperationError("Unkown model information.", ret)

    log_debug("Model of this node: %s" % (HDIConfig.Model.tostr(model)))
    return model

def get_remote_available_capacity(ipaddr, config, ssh_config):
    """ Get volume group free capacity.

    Get free capacity of volume group of a remote node.
    If multiple volume groups exist on the node, total free capacity is returned.

    Args:
        ipaddr (str): IP address of a remote node.
        config (Config.Config): Configuration of the scripts.
        ssh_config (Config.SSHConfig): SSH configuration.

    Returns:
        int: Total free capacity of the volume group(s) on the node.
    """
    print
    print "Getting available capacity on %s ..." % ipaddr

    # Doesn't specify REST credential as it is not used here
    cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(ipaddr), None)
    conn = HDIOperation.Connection(ipaddr, cred, config.retry_count, config.retry_interval, config.timeout)

    # Get Volume group information on the node
    vgxml = get_configs.get_vg(conn)
    try:
        vg_list = HDIConfig.parse_vg(xml_data=vgxml)
    except HDIConfig.FormatError, e:
        raise Exceptions.OperationError("Wrong volume group information was returned: %s" % str(e))

    log_debug("Volume groups on %s: %s" % (ipaddr, [(vg.volumeGroupName, vg.totalSizeInGigaByte, vg.freeSizeInGigaByte) for vg in vg_list]))

    # Calculate total free capacity
    free = 0
    for vg in vg_list:
        free += vg.freeSizeInGigaByte

    return free

def get_remote_available_lulist(ipaddr, config, ssh_config):
    """ Get a list of LUs which are not used on a remote node.

    Get a list of LUs which are not used on a remote node.

    Args:
        ipaddr (str): IP address of a remote node.
        config (Config.Config): Configuration of the scripts.
        ssh_config (Config.SSHConfig): SSH configuration.

    Returns:
        List of HDIConfig.LUInfo: List of available LUs.
    """
    print
    print "Getting a list of available LUs on %s ..." % ipaddr

    # Doesn't specify REST credential as it is not used here
    cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(ipaddr), None)
    conn = HDIOperation.Connection(ipaddr, cred, config.retry_count, config.retry_interval, config.timeout)

    # Get LU list on the node
    luxml = get_configs.get_lu(conn)
    try:
        lu_list = HDIConfig.parse_lus(xml_data=luxml)
    except HDIConfig.FormatError, e:
        raise Exceptions.OperationError("Wrong LU information was returned: %s" % str(e))

    log_debug("LUs on %s: %s" % (ipaddr, [lu.deviceFileName for lu in lu_list]))

    # Extract only 'Free' LUs
    free_lu_list = []
    for lu in lu_list:
        if lu.function == "Free":
            free_lu_list.append(lu)

    return free_lu_list

def get_remote_iplist(ipaddr, config, ssh_config):
    """ Get a list of IP addresses on a remote node.

    Args:
        ipaddr (str): IP address of a remote node.
        config (Config.Config): Configuration of the scripts.
        ssh_config (Config.SSHConfig): SSH configuration.

    Returns:
        List of str: List of IP addresses.
    """
    print
    print "Getting a list of IP addresses on %s ..." % ipaddr

    # Doesn't specify REST credential as it is not used here
    cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(ipaddr), None)
    conn = HDIOperation.Connection(ipaddr, cred, config.retry_count, config.retry_interval, config.timeout)

    # Get LU list on the node
    xml = get_configs.get_network(conn)
    try:
        iplist = HDIConfig.parse_interfaces(xml_data=xml)
    except HDIConfig.FormatError, e:
        raise Exceptions.OperationError("Wrong network information was returned: %s" % str(e))

    log_debug("IP addresses on %s: %s" % (ipaddr, iplist))

    return iplist

def is_running_healthily(config):
    """ Check if the node is running healthily.

    This function checks resource group status on this node
    (and the other node for cluster configuration) and checks
    if all resource groups are running without any errors and
    on primary nodes (i.e. not failed over).

    Args:
        config (Config.Config): Configuration of the scripts.
    Returns:
        bool: True if the resource groups are running healthily.
            False otherwise.
    """

    # No need to specify SSH credential because we are going to make a connection to localhost.
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)
    nodeop = HDIOperation.Node(conn)
    try:
        is_healthy = nodeop.is_running_healthily()
    except HDIOperation.OperationError, e:
        log_error("Failed to get resource group status. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get resource group status. error=%s" % e.reason, e.code)

    return is_healthy
