#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"

import sys
import os
import errno
import time

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception
import Utility
from Utility import print_status
import HDIConfig
import HDIClient
import HDIOperation
import FailoverConfig

sys.path.append(os.path.join(BASEDIR, "bin"))
import Exceptions
import Config

# Constants

lock_dir = 'lock'
register_lock_file = 'register_lock'
execute_lock_file = 'execute_lock'
alert_lock_file = 'alert_lock'

# Functions

def make_lock_dir_path():
    conf_dir_path = os.path.join(BASEDIR, 'conf')
    lock_dir_path = os.path.join(conf_dir_path, 'lock')

    try:
        os.mkdir(lock_dir_path)
    except EnvironmentError, e:
        if e.errno != errno.EEXIST:
            log_error("Failed to make a system directory %s. error=%s" % (lock_dir_path, str(e)))
            raise

    return lock_dir_path

def get_register_lock_file():
    """ Return lock file path for failover job management.

    Some necessary directory is created if it doesn't exist yet.

    Args:
        None
    Returns:
        str: File path of the lock file for failover job management.
    """

    lock_dir_path = make_lock_dir_path()
    return os.path.join(lock_dir_path, register_lock_file)

def get_execute_lock_file():
    """ Return lock file path for failover job execution.

    Args:
        None
    Returns:
        str: File path of the lock file for failover job execution.
    """

    lock_dir_path = make_lock_dir_path()
    return os.path.join(lock_dir_path, execute_lock_file)

def get_alert_lock_file():
    """ Return lock file path for sending an alert.

    Args:
        None
    Returns:
        str: File path of the lock file for sending an alert.
    """

    lock_dir_path = make_lock_dir_path()
    return os.path.join(lock_dir_path, alert_lock_file)

def check_execution_lock_on_the_other_node(config, wait=True):
    """ Check execution lock acquisition status on the other node in cluster.

    This function checks if execution lock is acquired on the other node in
    cluster, and return True if it is acquired.

    This function returns False when:
        - I am Single or VM node.
        - I am node 0 in cluster.
        - Execution lock is not acquired on the other node in cluster.

    Args:
        config (Config.Config): Configuration of the scripts.
        wait (bool): Specify False if this function is supposed to return
            immediately even if execution lock on the other node is acquired.
            If False is specified, this function will wait until execution
            lock on the other node is released.

    Returns:
        bool: True if execution lock is acquired on the other node in cluster.
            False otherwise.
            If 'wait' is True, always False will be returned.
    """
    log_info("Checking execution lock status on the other node ...")

    # Get ProcessingNode information.
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn_self = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)
    nodeop = HDIOperation.Node(conn_self)
    try:
        model = nodeop.get_model()
        node_number = nodeop.get_number()
        other_ip = nodeop.get_other_mng_ip()
    except HDIOperation.OperationError, e:
        log_error("Failed to get node information on the node. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get node information on the node. error=%s" % e.reason, e.code)
    log_info("model: %s, node_number: %s, other node: %s" % (HDIConfig.Model.tostr(model), node_number, other_ip))

    # Return False if I am Single or VM.
    if model == HDIConfig.Model.SINGLE:
        log_info("I am a Single or VM.")
        return False

    # Return False if I am node 0.
    if node_number == 0:
        log_info("I am node 0.")
        return False

    # I am node1 in cluster.
    # Get IP address of the other node.
    # Use management IP address to simplify this process.
    if other_ip is None:
        log_error("Could not get management IP address of the other node for some reason.")
        raise Exceptions.OperationError("Could not get management IP address of the other node.")

    # Wait for 30 seconds to avoid conflict
    SLEEP = 30
    print_status("Waiting %s seconds to avoid conflict with the other node in cluster ..." % SLEEP)
    time.sleep(SLEEP)

    # Is execution lock acquired on the other node?
    # Assume Unique SSH key is set up on the other node.
    # There must be a configuration group.
    confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    if len(confgroup_list) == 0:
        raise Exceptions.ConfigurationError("No configuration groups.")
    confgroup = confgroup_list[0]
    ssh_config = Config.SSHConfig(config.sys_conf_dir, confgroup=confgroup)
    cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(), None)
    conn_other = HDIOperation.Connection(other_ip, cred, config.retry_count, config.retry_interval, config.timeout)
    prog = "int/is_under_execution"
    cmd = "sudo maintexec %s" % os.path.join(BASEDIR, prog)
    op = HDIOperation.CommandOperation(cmd)
    while True:
        try:
            ret, out, err = op.execute(conn_other)
        except HDIOperation.ConnectionError, e:
            log_error("Failed to connect %s over SSH. error=%s" % (conn_other.host, str(e)))
            raise Exceptions.ConnectionError(e.reason, e.code)

        if ret == 0:
            locked = True
            log_info("Execution lock is acquired on the other node.")
        elif ret == 1:
            locked = False
            log_info("Execution lock is not acquired on the other node.")
        else:
            raise Exceptions.OperationError("An error occurred while checking execution lock status on %s. ret=%s, out=%s" % (conn_other.host, ret, out))

        if not wait or not locked:
            # Return immediately without waiting the lock being released
            break

        if locked:
            print_status("File system synchronization or failover execution is running on the other node. Waiting %s seconds and then will retry ..." % config.sleep_for_wait)
            time.sleep(config.sleep_for_wait)

    return locked
